#include "Estimator.h"

Filter filter_l, filter_r;
pthread_mutex_t odometryMutex;

bool Estimator::instanceFlag = false;
Estimator* Estimator::estimator = NULL;

Estimator* Estimator::getInstance(){
	if(instanceFlag==false){
		estimator = new Estimator();
		instanceFlag=true;
	}
	return estimator;
}

Estimator::Estimator(){
    robot.x_local = 0;          // estimated position in table coordinates
    robot.y_local = 0;
    robot.theta_local = 0;
    robot.u_body = 0;           // estimated velocity
    robot.w_body = 0;

    delta_body.r_body = 0;      // distance the robot has moved since the previous measurement
    delta_body.psi_body = 0;    // angle the robot has moved since the previous measurement
    delta_body.odo_left = 0;    // distance the odo.-wheels has moved since the previous measurement
    delta_body.odo_right = 0;

    odometry_left_prev = 0;     // previous odometry measurements
    odometry_right_prev = 0;
    theta_local_prev = 0;

    odo.left = 0;               // latest odometry measurements
    odo.right = 0;
    odo.timestamp = {0,0};
    odo_left_filtered = 0;
    odo_right_filtered = 0;

    opponent.x = 0;             // opponent position
    opponent.y = 0;

    odometryspeed_left = 0;     //velocity from odometry, along odometry wheels, in mm/s
    odometryspeed_right = 0;

    motor.motor2 = 0;           // output, left motor
    motor.motor1 = 0;           // output, right motor

    pthread_mutex_init (&odometryMutex , NULL);
    }

Estimator::~Estimator(){}

void Estimator::set(odometry_t odometry){

    pthread_mutex_lock(&odometryMutex);

    odo.timestamp = odometry.timestamp;

    odo.left = -odometry.left;  // Left odometry wheel is placed backwards
    odo.right = odometry.right;

    /* Filtering odometry measurements before estimating speed */
    odo_left_filtered = (double) odo.left;
    odo_right_filtered = (double) odo.right;

    filter_l.LPfilter(odo_left_filtered);
    filter_r.LPfilter(odo_right_filtered);

    pthread_mutex_unlock(&odometryMutex);

    // Number of encoder step since last measurement
    double odometry_delta_left = odo.left - odometry_left_prev;
	double odometry_delta_right = odo.right - odometry_right_prev;

	odometry_left_prev = odo.left;
	odometry_right_prev = odo.right;

    // Distance each wheel has moved in one time step. In mm.
	delta_body.odo_left = odometry_delta_left*ODOMETRY_DIAMETER_LEFT*M_PI/ODOMETRY_PULSES_PER_REVOLUTION;
	delta_body.odo_right = odometry_delta_right*ODOMETRY_DIAMETER_RIGHT*M_PI/ODOMETRY_PULSES_PER_REVOLUTION;

    // Distance and angle the robot has moved since last measurement
    delta_body.r_body = (delta_body.odo_left + delta_body.odo_right)/2;
	delta_body.psi_body = (delta_body.odo_left - delta_body.odo_right)/ODOMETRY_WHEEL_DISTANCE;

    /// Estimate of position and speed
    robot.x_local += delta_body.r_body * cos(robot.theta_local + delta_body.psi_body/2);
	robot.y_local += delta_body.r_body * sin(robot.theta_local + delta_body.psi_body/2);
	robot.theta_local += delta_body.psi_body;

	robot.u_body = delta_body.r_body/T;
	robot.w_body = delta_body.psi_body/T;

	odometryspeed_left = delta_body.odo_left/T;     // velocity along odometry wheels, in mm/s
    odometryspeed_right = delta_body.odo_right/T;
}

void Estimator::setPosition(double x, double y, double theta){
    robot.x_local = x;
    robot.y_local = y;
    robot.theta_local = theta;
}

void Estimator::setInit(robot_state r){
    robot.x_local = r.x_local;
    robot.y_local = r.y_local;
    robot.theta_local = r.theta_local;
    robot.u_body = r.u_body;
    robot.w_body = r.w_body;
}

void Estimator::setOpponentPosition(opponent_t opponent_info){
    double opponent_x = opponent_info.distance*cos(opponent_info.angle);
    double opponent_y = opponent_info.distance*sin(opponent_info.angle);
    opponent.x = robot.x_local + cos(robot.theta_local)*opponent_x - sin(robot.theta_local)*opponent_y;
    opponent.y = robot.y_local + cos(robot.theta_local)*opponent_y + sin(robot.theta_local)*opponent_x;
}

void Estimator::setMotorEstimate(motor_speed_t m){
    motor.motor2 = m.motor2;
    motor.motor1 = -m.motor1;
}

void Estimator::getMotorEstimate(motor_speed_t &m){
    m.motor2 = motor.motor2;
    m.motor1 = motor.motor1;
}

void Estimator::get(robot_state &rob){
    rob.x_local = robot.x_local;
    rob.y_local = robot.y_local;
    rob.theta_local = robot.theta_local;
    rob.u_body = robot.u_body;
    rob.w_body = robot.w_body;
}

void Estimator::get_delta(delta &speed){
    speed.r_body = delta_body.r_body;
    speed.psi_body = delta_body.psi_body;
    speed.odo_left = delta_body.odo_left;
    speed.odo_right = delta_body.odo_right;
}

void Estimator::get_odometry(odometry_t &o, double &odometryspeed_left, double &odometryspeed_right, double &lp_odo_l, double &lp_odo_r){
    pthread_mutex_lock(&odometryMutex);
    o.left = odo.left;                  // odometry measurement
    o.right = odo.right;
    o.timestamp = odo.timestamp;
    odometryspeed_left = this->odometryspeed_left;   // velocity from odometry
    odometryspeed_right = this->odometryspeed_right;
    lp_odo_l = odo_left_filtered;       // filtered odometry measurements
    lp_odo_r = odo_right_filtered;
    pthread_mutex_unlock(&odometryMutex);
}

void Estimator::getOpponentPosition(opponent_position &opponent){
    opponent.x = this->opponent.x;
    opponent.y = this->opponent.y;
}
