#include "communication.h"

using namespace std;

/** Global data structures **/
MessageQueue navsysQueue(PMQ_NAVSYS);
MessageQueue gatewayQueue(PMQ_GATEWAY);
MessageQueue aiQueue(PMQ_AI);
MessageQueue cvsysQueue(PMQ_CVSYS);

pthread_mutex_t startMutex;
odometry_t odometrydata;
pthread_t comThread;
motor_speed_t motorData;
odometry_t odometry_offset;
starting_position startpos;
waypoint_list list_of_waypoints;

Estimator *est = Estimator::getInstance();
RobotModel *rob = RobotModel::getInstance();

bool waypoints_received;
bool start_received;


void communicationInit(){
    /** Clear every queue **/
    navsysQueue.clear();
    gatewayQueue.clear();
    aiQueue.clear();

    waypoints_received = false;
    start_received = false;

    /*Create thread*/
    if(pthread_create(&comThread, NULL,communicationReceiveThread, NULL)){
        perror("communicationReceiveThread");
        cout << "Communication Receive Thread [error]" <<endl;
        exit(1);
    }
    else{
        cout << "Communication Receive Thread [ok]" << endl;
    }
}

void communicationExit(){
    navsysQueue.remove();
}

/** communicationReceiveThread **/
extern "C" void * communicationReceiveThread(void * arg){
    posix_message_t pm = {0};

    odometry_t odometry;
    robot_position robotposition;
    starting_position start;
    waypoint_list wp_list;
    motor_speed_t mSpeed;
    opponent_t opponent_pos;

    navsysQueue.clear();    //remove old messages in queue

    while(1){

        navsysQueue.receive(&pm);
        switch(pm.id)
        {
            case PM_NAVSYS_ODOMETRY:
                memcpy(&odometry,pm.data,sizeof(odometry_t));
                est->set(odometry);
                est->get(rob->robot_estimate);
                sendRobotPositionCvsys(odometry, &robotposition, rob->robot_estimate);
                //cout << "PM_NAVSYS_ODOMETRY  L: " << odometry.left << " R: " << odometry.right << endl;
                break;

            case PM_NAVSYS_MOTOR_SPEED:
                memcpy(&mSpeed,pm.data,sizeof(motor_speed_t));
                setMotorSpeed(mSpeed);
                est->setMotorEstimate(mSpeed);
                break;

            case PM_NAVSYS_STARTING_POSITION:
                memcpy(&start, pm.data, sizeof(starting_position));
                setStartingPosition(&start);
                pthread_mutex_lock(&startMutex);
                start_received = true;
                pthread_mutex_unlock(&startMutex);
                break;

            case PM_NAVSYS_UPDATE_WP:
                memcpy(&wp_list, pm.data, sizeof(waypoint_list));
                setWpList(&wp_list);
                break;

            case PM_NAVSYS_STOP:
                waypoints_received = false;
                rob->stopModel();
                cout << "stop command received" << endl;
                break;

            case PM_NAVSYS_OPPONENT:
                memcpy(&opponent_pos,pm.data,sizeof(opponent_t));
                est->setOpponentPosition(opponent_pos);
                est->getOpponentPosition(est->opponent);
                sendOpponentPositionCvsys(&est->opponent);
                break;

            /* case PM_NAVSYS_ABSOLUTE_POSITION:
                memcpy(&odometry,pm.data,sizeof(odometry_t));
            */

            default:
                cout << "MESSAGE NOT DEFINED "  << pm.id << endl;
                break;
        }
    }
    return NULL;
}

void resetOdometry(){
    odometry_offset = odometrydata;
    gatewayQueue.sendCommand(PM_SENSOR_ODOMETRY_RESET, PRIORITY_NORMAL);
}

void getMotorSpeed(motor_speed_t &ms){
    ms = motorData;
}

void setMotorSpeed(motor_speed_t &ms){
    motorData = ms;
}

void getStartingPosition(starting_position *s){
    *s = startpos;
}

void setStartingPosition(starting_position *s){
    startpos = *s;
}

void getWpList(waypoint_list *list){
    *list = list_of_waypoints;
}

void setWpList(waypoint_list *list){
    list_of_waypoints = *list;
    waypoints_received = true;
}

void sendPID(motor_pid_t *pid){
    gatewayQueue.sendStruct(PM_MOTOR_PID, PRIORITY_HIGH, sizeof(motor_pid_t), (char*) pid);
}

void sendMotorSpeed(motor_speed_ref_t *motorspeed){
    gatewayQueue.sendStruct(PM_MOTOR_SPEED_REF, PRIORITY_HIGH, sizeof(motor_speed_ref_t), (char*) motorspeed);
}

void sendRobotPositionAi(robot_position *robotposition){
    aiQueue.sendStruct(PM_AI_ROBOT_POSITION, PRIORITY_HIGH, sizeof(robot_position), (char*) robotposition);
}
void sendRobotPositionCvsys(odometry_t odo, robot_position *robotposition, robot_state robot){
    robotposition->x = robot.x_local;
    robotposition->y = robot.y_local;
    robotposition->psi = robot.theta_local;
    robotposition->speed = robot.u_body;
    robotposition->timestamp = odo.timestamp;
    cvsysQueue.sendStruct(PM_CVSYS_ROBOT_POSITION, PRIORITY_HIGH, sizeof(robot_position), (char*) robotposition);
}

void sendOpponentPositionAi(opponent_position *opponentposition){
    aiQueue.sendStruct(PM_AI_OPPONENT_POSITION, PRIORITY_HIGH, sizeof(opponent_position), (char*) opponentposition);
}
void sendOpponentPositionCvsys(opponent_position *opponentposition){
    cvsysQueue.sendStruct(PM_CVSYS_OPPONENT_POSITION, PRIORITY_HIGH, sizeof(opponent_position), (char*) opponentposition);
}

void sendWaypointReached(){
    aiQueue.sendCommand(PM_AI_WP_REACHED, PRIORITY_HIGH);
    waypoints_received = false;
}
