#include "pid.h"

using namespace std;

pthread_mutex_t pid_controller_mutex = PTHREAD_MUTEX_INITIALIZER;

void PID::init(){
    this->p_gain = 0.0;   /* 'P' proportional gain               */
    this->i_gain = 0.0;   /* 'I' integral gain                   */
    this->d_gain = 0.0;   /* 'D' derivative gain                 */

    this->fri_min = 0.0;  /* min-value for extra friction output */
    this->fri_max = 0.0;  /* max-value for extra friction output */
    this->setpt = 0.0;    /* set point                           */
    this->min = 0.0;      /* minimum output value                */
    this->max = 0.0;      /* maximum output value                */
    this->slew = 0.0;     /* maximum slew rate on output         */
    this->integral = 0.0;
    this->integral_max = PID_DEFAULT_INTEGRAL_MAX;
    this->integral_min = PID_DEFAULT_INTEGRAL_MIN;
    this->last_error = 0.0;
    this->last_output = 0.0;

    pthread_mutex_init(&setpoint_mutex, NULL);
}

PID::PID(double Kp, double Ki, double Kd, double slew, double max, double min){
    this->init();
    this->p_gain = Kp;
    this->i_gain = Ki;
    this->d_gain = Kd;
    this->min = min;
    this->max = max;
    this->slew = slew;
}

PID::~PID(){
    std::cout << "PID dead :'( \n";
}

void PID::setSetpoint(double setpt){
	pthread_mutex_lock(&setpoint_mutex); /* setpt-variabel */
    this->setpt = setpt;
    pthread_mutex_unlock(&setpoint_mutex); /* setpt-variabel */
}

/// Run PID-controller.  PV: value measured.
double PID::run(double PV){
    double     this_error;
    double     this_output; // computed output
    double     deriv;

    this_error = this->setpt - PV;
    //cout << " \t \t output: " << this_output << endl;

    deriv = (this_error - this->last_error);

    //integrator -- supposed to work only to overcome friction
    if (fabs(last_output)<=(fri_max+8) && fabs(last_output)>=(fri_max-8)) {
        this->integral += (this->setpt - PV);
    }
    else {
        this->integral = 0;
    }

    if( this->integral > this->integral_max) this->integral = this->integral_max;
    else if(this->integral < this->integral_min) this->integral = this->integral_min;

    /** PID **/
    this_output = 0.0
        + this->p_gain * this_error
        + this->i_gain * this->integral
        + this->d_gain * deriv;


    /** If friction to overcome to make the robot move.
        Decided by variables this->fri_max and this->fri_min **/
    //cout << " \t \t output1: " << this_output << endl;
    if ((fri_max!=0) && (fabs(this_output)<=fri_max)){
        // output from fri_max down to fri_min
        if (this_output>fri_min){
            this_output = fri_max;
        }
        // output from -fri_max up to -fri_min
        else if (this_output<-fri_min){
            this_output = -fri_max;
        }
        // from fri_min to 0 downwards, or from -fri_min to 0 upwards
        else if (this_output<=fri_min && this_output>=-fri_min) {
            this_output = 0;
        }
    }
    //cout << " \t \t output2: " << this_output << endl;

    /** check for slew rate limiting on the output change  */
    if ((this->slew != 0.0) && (fabs(last_output)>=fri_max)){
        if (this_output - this->last_output > this->slew){
            this_output = this->last_output + this->slew;
        }
        else if (this->last_output - this_output > this->slew){
            this_output = this->last_output - this->slew;
        }
    }
    cout << " \t \t output3: " << this_output << endl;

    /** check the output value for absolute limits */
    if (this_output < this->min){
        this_output = this->min;
    }
    else if (this_output > this->max){
        this_output = this->max;
    }

    /* store output and error for next iteration*/
    this->last_error = this_error;
    this->last_output = this_output;
    //cout << " \t \t output4: " << this_output << endl;
    return (this_output);
}

void* motor_speed_thread(void* args){
	struct timeval time_end, time_start;
	int sleeptime_us = 1000000 / ((struct pid_thread_args*) args)->frequency;
	PID *pid_1 = ((struct pid_thread_args*) args)->pid_1,
		*pid_2 = ((struct pid_thread_args*) args)->pid_2;
	output_speed* output = ((struct pid_thread_args*) args)->output;
	bool dist_controller = ((struct pid_thread_args*) args)->dist_controller;

	while(1){
		gettimeofday(&time_start, NULL);

        /** Kjører posisjonsregulatorene **/
        pthread_mutex_lock(&pid_controller_mutex); // Protect output-variable
        if (dist_controller == true) { // run both translation and rotation
            output->u_body = round(pid_1->run(0));
            output->w_body = round(pid_2->run(0));
        }
        else { // dist_controller == false
            output->u_body = 0;
            output->w_body = round(pid_2->run(0));
        }
        pthread_mutex_unlock(&pid_controller_mutex); // Beskytter output-variabel

		/* Sleep to get a fixed update frequency */
        gettimeofday(&time_end, NULL);
        usleep(sleeptime_us - (time_end.tv_usec - time_start.tv_usec));
	}
	pthread_exit(NULL);
}
